#include <ATen/Tensor.h>
#include <ATen/core/dispatch/Dispatcher.h>

// ${generated_comment}
// NOTE See [Sharded File] comment in VariableType

#ifndef AT_PER_OPERATOR_HEADERS
#include <ATen/Operators.h>
#else
${operator_headers}
#endif

${static_dispatch_extra_headers}

namespace at { namespace _ops {

${definitions}

}} // namespace at::_ops
