#pragma once

// @generated from tools/autograd/templates/python_functions.h

// Python bindings for automatically generated autograd functions

namespace torch { namespace autograd { namespace generated {

void initialize_autogenerated_functions_0();
void initialize_autogenerated_functions_1();
void initialize_autogenerated_functions_2();
void initialize_autogenerated_functions_3();
void initialize_autogenerated_functions_4();

inline void initialize_autogenerated_functions() {
  initialize_autogenerated_functions_0();
  initialize_autogenerated_functions_1();
  initialize_autogenerated_functions_2();
  initialize_autogenerated_functions_3();
  initialize_autogenerated_functions_4();
}

}}} // namespace torch::autograd::generated
