# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: caffe2/proto/torch.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from caffe2.proto import caffe2_pb2 as caffe2_dot_proto_dot_caffe2__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x63\x61\x66\x66\x65\x32/proto/torch.proto\x12\x05torch\x1a\x19\x63\x61\x66\x66\x65\x32/proto/caffe2.proto\"\x18\n\tRecordRef\x12\x0b\n\x03key\x18\x01 \x01(\t\"\xeb\x01\n\tTensorDef\x12\x0c\n\x04\x64ims\x18\x01 \x03(\x03\x12\x0e\n\x06offset\x18\x02 \x01(\x03\x12\x0f\n\x07strides\x18\x03 \x03(\x03\x12\x15\n\rrequires_grad\x18\x04 \x01(\x08\x12/\n\tdata_type\x18\x05 \x01(\x0e\x32\x1c.caffe2.TensorProto.DataType\x12\x1e\n\x04\x64\x61ta\x18\x06 \x01(\x0b\x32\x10.torch.RecordRef\x12\x0e\n\x06\x64\x65vice\x18\x07 \x01(\t\x12\x14\n\x0cis_quantized\x18\x08 \x01(\x08\x12\r\n\x05scale\x18\t \x01(\x01\x12\x12\n\nzero_point\x18\n \x01(\x03\"6\n\x0c\x41ttributeDef\x12\x0c\n\x04type\x18\x01 \x02(\t\x12\x0c\n\x04name\x18\x02 \x02(\t\x12\n\n\x02id\x18\x03 \x02(\x03\"B\n\x0cParameterDef\x12\x11\n\tis_buffer\x18\x01 \x01(\x08\x12\x11\n\ttensor_id\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\"\x95\x03\n\tModuleDef\x12$\n\nsubmodules\x18\x01 \x03(\x0b\x32\x10.torch.ModuleDef\x12+\n\x11torchscript_arena\x18\x02 \x01(\x0b\x32\x10.torch.RecordRef\x12#\n\x0b\x63\x61\x66\x66\x65\x32_nets\x18\x03 \x03(\x0b\x32\x0e.caffe2.NetDef\x12&\n\x0cpickle_arena\x18\x04 \x01(\x0b\x32\x10.torch.RecordRef\x12#\n\tcpp_arena\x18\x05 \x01(\x0b\x32\x10.torch.RecordRef\x12\'\n\nparameters\x18\x06 \x03(\x0b\x32\x13.torch.ParameterDef\x12\x0c\n\x04name\x18\x07 \x01(\t\x12\x10\n\x08optimize\x18\x08 \x01(\x08\x12\'\n\nattributes\x18\t \x03(\x0b\x32\x13.torch.AttributeDef\x12\x1e\n\x16get_state_attribute_id\x18\n \x01(\x03\x12\x31\n\x17torchscript_debug_arena\x18\x0b \x01(\x0b\x32\x10.torch.RecordRef\"5\n\x06LibDef\x12+\n\x11torchscript_arena\x18\x01 \x01(\x0b\x32\x10.torch.RecordRef\"\xa8\x01\n\x08ModelDef\x12\x15\n\rproto_version\x18\x01 \x01(\x03\x12%\n\x0bmain_module\x18\x02 \x01(\x0b\x32\x10.torch.ModuleDef\x12\x15\n\rproducer_name\x18\x03 \x01(\t\x12\x18\n\x10producer_version\x18\x04 \x01(\t\x12!\n\x07tensors\x18\x05 \x03(\x0b\x32\x10.torch.TensorDefJ\x04\x08\t\x10\nR\x04libs*(\n\x0cProtoVersion\x12\x18\n\x14PROTO_VERSION_NEWEST\x10\x06')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'caffe2.proto.torch_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PROTOVERSION._serialized_start=1084
  _PROTOVERSION._serialized_end=1124
  _RECORDREF._serialized_start=62
  _RECORDREF._serialized_end=86
  _TENSORDEF._serialized_start=89
  _TENSORDEF._serialized_end=324
  _ATTRIBUTEDEF._serialized_start=326
  _ATTRIBUTEDEF._serialized_end=380
  _PARAMETERDEF._serialized_start=382
  _PARAMETERDEF._serialized_end=448
  _MODULEDEF._serialized_start=451
  _MODULEDEF._serialized_end=856
  _LIBDEF._serialized_start=858
  _LIBDEF._serialized_end=911
  _MODELDEF._serialized_start=914
  _MODELDEF._serialized_end=1082
# @@protoc_insertion_point(module_scope)
